@@ -7,6 +7,8 @@
  *
  */
 
+
+#include <byteswap.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
@@ -34,6 +36,11 @@
 /* libpcap */
 #include <pcap.h>
 #include "thc-ipv6.h"
+/* libssl */
+#include <openssl/evp.h>
+#include <openssl/sha.h>
+
+
 
 int debug = 0;
 
@@ -1493,3 +1500,330 @@ unsigned char *thc_memstr(char *haystack, char *needle, int haystack_length,
              return (haystack + i);
     return NULL;
 }
+
+
+/*
+ * Added by willdamn <willdamn@gmail.com>
+ * 2006/07
+ */
+
+
+thc_key_t *thc_generate_key(int key_len)
+{
+   thc_key_t *key;
+   
+   if((key = (thc_key_t *)malloc(sizeof(thc_key_t))) == NULL)
+       return NULL;
+             
+   if((key->rsa = RSA_generate_key(key_len, 65535, NULL, NULL)) == NULL)
+       return NULL;
+   key->len = key_len;
+                            
+   return key;
+}
+
+
+thc_cga_hdr *thc_generate_cga(unsigned char *prefix, thc_key_t *key, unsigned char **cga)
+{
+   thc_cga_hdr *cga_hdr;
+   unsigned char md_value[EVP_MAX_MD_SIZE];
+   unsigned char *p, *tmp;
+   int klen, rand_fd, cgasize;
+
+
+    if((cga_hdr = (thc_cga_hdr *)malloc(sizeof(thc_cga_hdr))) == NULL)
+        return NULL;
+    
+    cga_hdr->type = 11;
+    
+    /* 
+     * prepare CGA paramater
+     */
+    
+    /* CGA header & mod_value, prefix, collision_count from CGA parameter*/
+    cgasize = 29;
+    
+    if((rand_fd = open("/dev/urandom", O_RDONLY)) < 0 )
+        { printf("Cannot open source of randomness!\n"); return NULL; }
+    read(rand_fd, cga_hdr->modifier, 16);
+    close(rand_fd);
+
+    /* DER-encode public key */
+    klen = i2d_RSA_PUBKEY(key->rsa, NULL);
+    if((cga_hdr->pub_key = (unsigned char *)malloc(klen)) == NULL)
+        return NULL;
+    p = cga_hdr->pub_key;
+    klen = i2d_RSA_PUBKEY(key->rsa, &p); 
+
+    key->len = klen;
+    cgasize += klen;
+        
+    cga_hdr->collision_cnt = 0;    
+    memcpy(cga_hdr->prefix, prefix, 8); 
+    
+    if((tmp = malloc(cgasize - 4)) == NULL)
+      { perror("tmp malloc "); return NULL; }
+    
+    memcpy(tmp, cga_hdr->modifier, 25);
+    memcpy(tmp+25, cga_hdr->pub_key, klen);
+    
+    /* compute hash1 */
+    SHA1(tmp, cgasize - 4, md_value);
+    free(tmp);
+    
+    if(cgasize % 8 == 0)
+       {
+         cga_hdr->len = cgasize / 8;
+         cga_hdr->pad_len = 0;
+       }
+    else
+       {
+         cga_hdr->len = cgasize / 8 + 1;
+         cga_hdr->pad_len = cga_hdr->len * 8 - cgasize; 
+         cga_hdr->pad = (char *)malloc(cga_hdr->pad_len);
+       }
+            
+    /* 
+     * Prepare CGA 
+     */
+    
+    if((*cga = (char *)malloc(16)) == NULL) 
+        return NULL;
+   
+    memcpy(*cga, prefix, 8);
+    /* add address identifier to cga */
+    memcpy(*cga+8, md_value, 8);
+    /* set "U" & "G" bits ; currently sec equals 0*/
+    *(*cga+8) &= 0x1c;
+                                       
+    return cga_hdr;                                    
+}
+
+thc_timestamp_hdr *generate_timestamp(void)
+{
+    thc_timestamp_hdr *timestamp;
+    struct timeval time;
+
+    if((timestamp = (thc_timestamp_hdr *)calloc(1, sizeof(thc_timestamp_hdr))) == NULL)
+        return NULL;
+    
+    timestamp->type = 13;
+    timestamp->len = 2;
+    
+    gettimeofday(&time, NULL);
+    timestamp->timeval = bswap_64(time.tv_sec << 16);
+
+    return timestamp;
+}
+
+thc_nonce_hdr *generate_nonce(void)
+{
+    thc_nonce_hdr *nonce;
+    
+    if((nonce = (thc_nonce_hdr *)malloc(sizeof(thc_nonce_hdr))) == NULL)
+      return NULL;
+    
+    nonce->type = 14;
+    nonce->nonce[0] = nonce->nonce[3] = 0xa; 
+    nonce->nonce[1] = nonce->nonce[4] = 0xc; 
+    nonce->nonce[2] = nonce->nonce[5] = 0xe;
+    nonce->len = sizeof(thc_nonce_hdr) / 8;
+
+    return nonce;    
+}
+
+thc_rsa_hdr *thc_generate_rsa(char *data2sign, int data2sign_len, thc_cga_hdr *cga_hdr, thc_key_t *key)
+{
+   thc_rsa_hdr *rsa_hdr;
+   unsigned char md_value[EVP_MAX_MD_SIZE], hash[20];
+   int rsa_hdr_len, sign_len, fd;
+         
+   if((rsa_hdr = (thc_rsa_hdr *)malloc(sizeof(thc_rsa_hdr))) == NULL)
+     return NULL;
+       
+   rsa_hdr->type = 12;
+                    
+   /* compute public key hash */
+   SHA1(cga_hdr->pub_key, key->len, md_value);
+   memcpy(rsa_hdr->key_hash, md_value, 16);
+  
+   /* If cga type tag's unknown set a bad RSA signature, e.g useful for DoS */
+
+   if(data2sign_len > 0)
+       SHA1(data2sign, data2sign_len, hash);
+   else
+     {
+       fd = open("/dev/urandom", O_RDONLY);
+       read(fd, hash, 20);
+       close(fd);
+     }
+
+   sign_len = RSA_size(key->rsa);
+ 
+   if((rsa_hdr->sign = malloc(sign_len)) == NULL)
+      { free(rsa_hdr); return NULL; }
+ 
+   if(RSA_sign(NID_sha1, hash, 20, rsa_hdr->sign, &sign_len, key->rsa) == 0)
+      { printf("Error during generating RSA signature! \n"); return NULL; }
+ 
+    rsa_hdr_len = 20 + sign_len;
+ 
+   if(rsa_hdr_len % 8 == 0)
+      {
+        rsa_hdr->len = rsa_hdr_len / 8;
+        rsa_hdr->pad = NULL;
+      }
+   else
+      {
+        rsa_hdr->len = rsa_hdr_len / 8 + 1;
+        rsa_hdr->pad = malloc(rsa_hdr->len * 8 - rsa_hdr_len);
+      }
+   return rsa_hdr;
+ }
+ 
+ 
+int thc_add_send(unsigned char *pkt, int *pkt_len, int type, int code, unsigned int flags,
+		 unsigned char *data, int data_len, thc_cga_hdr *cga_hdr, thc_key_t *key, unsigned char *tag, int checksum) 
+{ 
+   thc_ipv6_hdr *hdr = (thc_ipv6_hdr *) pkt;
+   thc_icmp6_hdr *ihdr = malloc(sizeof(thc_icmp6_hdr));
+   thc_ipv6_ext_hdr *ehdr;
+   thc_nonce_hdr *nonce_hdr = NULL;
+   thc_timestamp_hdr *timestamp_hdr = NULL;
+   thc_rsa_hdr *rsa_hdr = NULL;
+   unsigned char *ndp_opt_buff, *data2sign = NULL;
+   char *buff;
+   int ndp_opt_len, data2sign_len, offset;
+   
+   /* build standard part of ND message */
+   
+   if (ihdr == NULL)
+       return -1;
+   memset(ihdr, 0, sizeof(thc_icmp6_hdr));
+ 
+   if (hdr->final != NULL) {
+     ehdr = (thc_ipv6_ext_hdr *) hdr->final;
+     ehdr->next_segment = (char *) ihdr;
+     ehdr->next=NXT_ICMP6;
+   } else {
+     hdr->next_segment = (char *) ihdr;
+     hdr->next=NXT_ICMP6;
+   }
+   hdr->final = (char *) ihdr;
+   hdr->final_type=NXT_ICMP6;
+ 
+   ihdr->type = type;
+   ihdr->code = code;
+   ihdr->flags = flags;
+ 
+   if (checksum == 0) {
+     ihdr->checksum = DO_CHECKSUM;
+   } else
+     ihdr->checksum = checksum;
+     
+   if (data_len > 0 && data != NULL)
+       ndp_opt_len = data_len;
+   else 
+       ndp_opt_len = 0;
+                           
+   hdr->length += 8;
+   *pkt_len += 8;
+             
+   /* 
+    * add various security features to ND message
+    */
+   
+   /* determine options' total length */
+   
+   if((cga_hdr == NULL))
+       return -1;
+   
+   ndp_opt_len += cga_hdr->len * 8;   
+       
+   if((timestamp_hdr = generate_timestamp()) == NULL)
+       return -1;  
+   ndp_opt_len += timestamp_hdr->len * 8;
+   
+   if((nonce_hdr = generate_nonce()) == NULL)
+      { free(timestamp_hdr); return -1; }
+   ndp_opt_len += nonce_hdr->len * 8; 
+   
+   
+   /* create options buffer */
+          
+   if((ndp_opt_buff = (char *)malloc(ndp_opt_len)) == NULL)
+      { free(timestamp_hdr); free(nonce_hdr); return -1; } 
+  
+   offset = 0;
+     
+   if(data != NULL) 
+     {
+       memcpy(ndp_opt_buff+offset, data, data_len);
+       offset += data_len;
+     }
+   
+   /* CGA option */
+   memcpy(ndp_opt_buff+offset, cga_hdr, 29);
+   memcpy(ndp_opt_buff+offset+29, cga_hdr->pub_key, key->len);
+   offset += (cga_hdr->len * 8);
+   /* timestamp option */
+   memcpy(ndp_opt_buff+offset, timestamp_hdr, timestamp_hdr->len * 8);
+   offset += timestamp_hdr->len * 8;
+   free(timestamp_hdr);
+   /* nonce option */
+   memcpy(ndp_opt_buff+offset, nonce_hdr, nonce_hdr->len * 8);
+   offset += nonce_hdr->len * 8;
+   free(nonce_hdr);
+      
+   /* 
+    * RSA signature 
+    *   
+    * If CGA message type tag given compute correct RSA signature
+    * otherwise set option with incorrect one 
+    */
+    
+   if(tag != NULL)
+     {
+       data2sign_len = 52 + ndp_opt_len;
+       if((data2sign = (char *)malloc(data2sign_len)) == NULL)
+         { free(ndp_opt_buff); return -1; }
+       memcpy(data2sign, tag, 16);
+       memcpy(data2sign+16, hdr->src, 16);
+       memcpy(data2sign+32, hdr->dst, 16);
+       
+       /* compute icmp checksum that is needed to compute rsa signature */
+       if((buff=malloc(8 + ndp_opt_len)) == NULL)
+         { free(data2sign); free(ndp_opt_buff); return -1; } 
+       memcpy(buff, ihdr, 8);
+       memcpy(buff+8, ndp_opt_buff, ndp_opt_len);
+       ihdr->checksum = checksum_pseudo_header(hdr->src, hdr->dst, NXT_ICMP6, buff, 8 + ndp_opt_len);
+       free(buff);
+       memcpy(data2sign+48, &ihdr->type, 4);
+       ihdr->checksum = 0;
+       
+       memcpy(data2sign+52, ndp_opt_buff, ndp_opt_len);
+     }
+   else
+       data2sign_len = -1;
+      
+   if((rsa_hdr = thc_generate_rsa(data2sign, data2sign_len, cga_hdr, key)) == NULL)
+       { free(ndp_opt_buff); return -1; }
+   ihdr->data_len = ndp_opt_len + rsa_hdr->len * 8;
+   
+   free(data2sign);
+   
+   /* create 'real' buffer for NDP options */ 
+   if((ihdr->data = (unsigned char *)malloc(ihdr->data_len)) == NULL)
+       { free(ndp_opt_buff); free(rsa_hdr); return -1; }
+   memcpy(ihdr->data, ndp_opt_buff, ndp_opt_len);
+   free(ndp_opt_buff);
+   
+   /* RSA signature option */
+   memcpy(ihdr->data+ndp_opt_len, rsa_hdr, 20);
+   memcpy(ihdr->data+ndp_opt_len+20, rsa_hdr->sign, rsa_hdr->len * 8 - 20);
+    
+   hdr->length += ihdr->data_len;
+   *pkt_len += ihdr->data_len;
+   
+   return 0; 
+}
