@@ -10,8 +10,9 @@
 #define _THC_IPV6_H
 
 #include <pcap.h>
+#include <openssl/rsa.h>
 
-#define VERSION 	"v0.6"
+#define VERSION 	"v0.7"
 #define AUTHOR 		"van Hauser / THC <vh@thc.org>"
 #define RESOURCE	"www.thc.org"
 
@@ -89,6 +90,7 @@ unsigned char *thc_string2ipv6(unsigned char *string);
 unsigned char *thc_string2notation(unsigned char *string);
 unsigned char *thc_memstr(char *haystack, char *needle, int haystack_length, int needle_length);
 
+
 #define DO_CHECKSUM 0xffff
 
 #define NXT_IP6 41
@@ -187,5 +189,58 @@ typedef struct {
         unsigned int           nl_groups;
 } sockaddr_nl;
 */
+
+typedef struct {
+  unsigned char type;
+  unsigned char len;
+  unsigned char pad_len;
+  unsigned char resv;
+/* cga params */
+  unsigned char modifier[16];
+  unsigned char prefix[8];
+  unsigned char collision_cnt;
+  unsigned char coll2;
+  unsigned char *pub_key; 
+  unsigned char *exts;
+/* end of cga params */
+  unsigned char *pad;
+} thc_cga_hdr;
+
+typedef struct {
+  unsigned char type;
+  unsigned char len;
+  unsigned char resv[6];
+  unsigned long long timeval;
+} thc_timestamp_hdr;
+
+typedef struct {
+  unsigned char type;
+  unsigned char len;
+  char nonce[6];
+} thc_nonce_hdr;
+
+typedef struct {
+  unsigned char type;
+  unsigned char len;
+  short int resv;
+  unsigned char key_hash[16];
+  char *sign;
+  char *pad;
+} thc_rsa_hdr;       
+
+typedef struct {
+  RSA *rsa;
+  int len;
+} thc_key_t;
+
+typedef struct {
+  unsigned char *data;
+  /* DER-encoded key length */
+  int len;
+} opt_t;
+
+extern thc_key_t *thc_generate_key(int key_len);
+extern thc_cga_hdr *thc_generate_cga(unsigned char *prefix, thc_key_t *key, unsigned char **cga);
+
 #endif
 
